

# Dataset name mapping - lowercase underscore to official name
DATASET_NAME_MAP = {
    'sec_accessions': 'SEC Accessions Master Index',
    'sec_master_submissions': 'SEC Master Submissions Table',
    'sec_accession_cik_table': 'SEC Accession CIK Table',
    'sec_documents_table': 'SEC Documents Table',
    'sec_submission_details_table': 'SEC Submissions Details Table',
    'simple_xbrl_table': 'Simple XBRL Table',
    'proxy_voting_records_table': 'Proxy Voting Records Table',
    'institutional_holdings_table': 'Institutional Holdings Table',
    'metadata_ownership_table': 'Insider Ownership Metadata Table',
    'reporting_owner_ownership_table': 'Insider Reporting Owner Table',
    'non_derivative_transaction_ownership_table': 'Insider Non-Derivative Transactions Table',
    'non_derivative_holding_ownership_table': 'Insider Non-Derivative Holdings Table',
    'derivative_transaction_ownership_table': 'Insider Derivative Transactions Table',
    'derivative_holding_ownership_table': 'Insider Derivative Holdings Table',
    'owner_signature_ownership_table': 'Insider Owner Signatures Table',
}



import os

def get_api_key():
    key = os.environ.get('DATAMULE_API_KEY')
    if not key:
        raise EnvironmentError("DATAMULE_API_KEY environment variable is not set.")
    return key


api_key = get_api_key()


import asyncio
import aiohttp
import ssl
import time
from tqdm import tqdm
from .api_key import api_key as API_KEY

_total_cost = 0
_remaining_balance = None

async def _fetch_page(session, database, params, page=1, page_size=25000):
    url = f"https://api.datamule.xyz/{database}"
    
    query_params = params.copy()
    query_params["page"] = page
    query_params["pageSize"] = page_size
    query_params["api_key"] = API_KEY
    
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
    }
    
    async with session.get(url, params=query_params, headers=headers) as response:
        data = await response.json()
        if not data.get('success'):
            raise ValueError(f"API request failed: {data.get('error')}")
        
        billing = data.get('metadata', {}).get('billing', {})
        page_cost = billing.get('total_charge', 0)
        remaining_balance = billing.get('remaining_balance')
        
        pagination = data.get('metadata', {}).get('pagination', {})
        
        result_data = data.get('data', [])
        if isinstance(result_data, dict) and 'data' in result_data:
            result_data = result_data['data']
        
        return result_data, pagination, page_cost, remaining_balance


async def _execute_query(database, **kwargs):
    global _total_cost, _remaining_balance
    _total_cost = 0

    page_size = kwargs.pop('page_size', 25000)
    quiet = kwargs.pop('quiet', False)

    params = {}
    for key, value in kwargs.items():
        if value is None:
            continue
        elif isinstance(value, list):
            params[key] = ','.join([str(val) for val in value])
        elif isinstance(value, tuple):
            params[f"{key}_START"] = value[0]
            params[f"{key}_END"] = value[1]
        else:
            params[key] = value

    start_time = time.time()
    total_items = 0
    pages_processed = 0
    results = []

    connector = aiohttp.TCPConnector(ssl=ssl.create_default_context())
    async with aiohttp.ClientSession(connector=connector) as session:
        if not quiet:
            pbar = tqdm(unit="page", bar_format="{desc}: {n_fmt} {unit} [{elapsed}<{remaining}, {rate_fmt}{postfix}]")
            pbar.set_description("Fetching data")

        current_page = 1
        has_more = True

        while has_more:
            page_results, pagination, page_cost, _remaining_balance = await _fetch_page(
                session, database, params, page=current_page, page_size=page_size
            )

            results.extend(page_results)
            _total_cost += page_cost
            pages_processed += 1
            total_items += len(page_results)

            if not quiet:
                pbar.set_description(f"Fetching data (page {current_page})")
                pbar.set_postfix_str(f"cost=${_total_cost:.4f} | balance=${_remaining_balance:.2f}")
                pbar.update(1)

            has_more = pagination.get('hasMore', False)
            current_page += 1

            if pages_processed == 1 and not quiet:
                records_per_page = pagination.get('currentPageRecords', len(page_results))
                if records_per_page > 0:
                    pbar.write(f"Retrieved {records_per_page} records (page 1) - Fetching additional pages...")
                else:
                    pbar.write("No records found matching criteria")
                    break

        if not quiet:
            pbar.close()
            elapsed_time = time.time() - start_time
            print("\nQuery complete:")
            print(f"- Retrieved {total_items} records across {pages_processed} pages")
            print(f"- Total cost: ${_total_cost:.4f}")
            print(f"- Remaining balance: ${_remaining_balance:.2f}")
            print(f"- Time: {elapsed_time:.1f} seconds")

    return results


def query_database(database, **kwargs):
    return asyncio.run(_execute_query(database=database, **kwargs))





import asyncio
import aiohttp
import aioboto3
import ssl
import json
import urllib
from tqdm import tqdm
from datetime import datetime, timezone

from ...api_key import api_key
from ...datasets import DATASET_NAME_MAP


async def _get_dataset_url(session, dataset_name):
    api_url = f"https://api.datamule.xyz/dataset/{urllib.parse.quote(dataset_name)}?api_key={api_key}"
    async with session.get(api_url, headers={'User-Agent': 'datamule-python'}) as response:
        data = await response.json()
    if not data.get('success'):
        raise Exception(f"API error: {data.get('error', 'Unknown error')}")
    billing = data.get('metadata', {}).get('billing', {})
    return data['data']['download_url'], data['data']['size_gb'], billing


async def _transfer_dataset(session, s3_client, semaphore, dataset, bucket, prefix=None, retry_errors=3, multipart_threshold_mb=100, chunk_size_mb=8):
    dataset_name = DATASET_NAME_MAP.get(dataset)
    if not dataset_name:
        return {'success': False, 'dataset': dataset, 'error': f"Unknown dataset: {dataset}"}

    download_url, size_gb, billing = await _get_dataset_url(session, dataset_name)

    parsed = urllib.parse.urlparse(download_url)
    path = urllib.parse.parse_qs(parsed.query).get('path', [''])[0]
    key = urllib.parse.unquote(path.split('/')[-1]) or f"{dataset}.download"
    if prefix:
        key = f"{prefix.rstrip('/')}/{key}"

    use_multipart = size_gb * 1024 > multipart_threshold_mb
    chunk_size = chunk_size_mb * 1024 * 1024

    async with semaphore:
        last_error = None
        for attempt in range(retry_errors + 1):
            mpu_id = None
            try:
                async with session.get(download_url) as response:
                    if response.status != 200:
                        raise aiohttp.ClientResponseError(response.request_info, response.history, status=response.status)

                    if use_multipart:
                        mpu = await s3_client.create_multipart_upload(
                            Bucket=bucket, Key=key,
                            ContentType=response.headers.get('Content-Type', 'application/octet-stream'),
                            Metadata={'source-url': download_url, 'transfer-date': datetime.now(timezone.utc).isoformat()}
                        )
                        mpu_id = mpu['UploadId']
                        parts = []
                        part_number = 1
                        buffer = b""

                        async for chunk in response.content.iter_chunked(chunk_size):
                            buffer += chunk
                            if len(buffer) >= chunk_size:
                                part = await s3_client.upload_part(
                                    Bucket=bucket, Key=key,
                                    PartNumber=part_number, UploadId=mpu_id, Body=buffer
                                )
                                parts.append({'PartNumber': part_number, 'ETag': part['ETag']})
                                part_number += 1
                                buffer = b""

                        if buffer:
                            part = await s3_client.upload_part(
                                Bucket=bucket, Key=key,
                                PartNumber=part_number, UploadId=mpu_id, Body=buffer
                            )
                            parts.append({'PartNumber': part_number, 'ETag': part['ETag']})

                        await s3_client.complete_multipart_upload(
                            Bucket=bucket, Key=key,
                            UploadId=mpu_id, MultipartUpload={'Parts': parts}
                        )

                    else:
                        content = await response.read()
                        await s3_client.put_object(
                            Bucket=bucket, Key=key, Body=content,
                            ContentType=response.headers.get('Content-Type', 'application/octet-stream'),
                            Metadata={'source-url': download_url, 'transfer-date': datetime.now(timezone.utc).isoformat()}
                        )

                    return {'success': True, 'dataset': dataset, 'size_bytes': int(size_gb * 1024**3), 'billing': billing}

            except Exception as e:
                if mpu_id:
                    try:
                        await s3_client.abort_multipart_upload(Bucket=bucket, Key=key, UploadId=mpu_id)
                    except Exception:
                        pass
                if attempt < retry_errors:
                    download_url, _, _ = await _get_dataset_url(session, dataset_name)
                    await asyncio.sleep(2 ** attempt)
                last_error = e

        return {'success': False, 'dataset': dataset, 'error': str(last_error)}


async def _transfer_datasets(datasets, s3_credentials, max_workers, retry_errors, prefix=None):
    connector = aiohttp.TCPConnector(limit=max_workers, ssl=ssl.create_default_context())
    async with aiohttp.ClientSession(connector=connector, timeout=aiohttp.ClientTimeout(total=600)) as session:
        async with aioboto3.Session().client(
            's3',
            aws_access_key_id=s3_credentials['aws_access_key_id'],
            aws_secret_access_key=s3_credentials['aws_secret_access_key'],
            region_name=s3_credentials['region_name']
        ) as s3_client:
            semaphore = asyncio.Semaphore(max_workers)
            tasks = [_transfer_dataset(session, s3_client, semaphore, d, s3_credentials['bucket_name'], prefix, retry_errors) for d in datasets]

            failed, total_bytes = [], 0
            with tqdm(total=len(datasets), desc="Transferring datasets", unit="dataset") as pbar:
                for coro in asyncio.as_completed(tasks):
                    result = await coro
                    if result['success']:
                        total_bytes += result['size_bytes']
                    else:
                        failed.append(result)
                    pbar.set_postfix({'Total': f"{total_bytes / (1024**3):.2f} GB"})
                    pbar.update(1)

            return failed


def datasets_transfer(datasets, s3_credentials, max_workers=4, errors_json_filename='errors_datasets_transfer.json', retry_errors=3, prefix=None):
    failed = asyncio.run(_transfer_datasets(datasets, s3_credentials, max_workers, retry_errors, prefix=prefix))

    if failed and errors_json_filename:
        with open(errors_json_filename, 'w') as f:
            json.dump(failed, f, indent=2)
        print(f"Saved {len(failed)} errors to {errors_json_filename}")

    print(f"Transfer complete: {len(datasets) - len(failed)}/{len(datasets)} datasets successful")


from datetime import datetime, timedelta
from ..query_database import query_database
from ..utils.format_accession import format_accession

def _generate_dates(filing_date):
    if isinstance(filing_date, str):
        return [filing_date]
    elif isinstance(filing_date, list):
        return filing_date
    elif isinstance(filing_date, tuple):
        start = datetime.strptime(filing_date[0], '%Y-%m-%d')
        end = datetime.strptime(filing_date[1], '%Y-%m-%d')
        dates = []
        current = start
        while current <= end:
            dates.append(current.strftime('%Y-%m-%d'))
            current += timedelta(days=1)
        return dates
    raise ValueError('filing_date must be a string, list, or (start, end) tuple')

def _get_urls(submission_type=None, cik=None, filing_date=None, accession_number=None):
    results = query_database('sec-filings-lookup',
        cik=cik,
        submissionType=submission_type,
        filingDate=filing_date,
        accessionNumber=accession_number
    )
    return [f"https://sec-library.datamule.xyz/{format_accession(result['accessionNumber'], 'no-dash')}.sgml" for result in results]


import urllib.request
import urllib.parse
from tqdm import tqdm
import json

from .datasets import DATASET_NAME_MAP
from .api_key import api_key

def download_dataset(dataset, filename=None):
    # Map dataset name to official name
    dataset_name = DATASET_NAME_MAP.get(dataset)
    if not dataset_name:
        raise ValueError(f"Unknown dataset: {dataset}")
    
    # Get download URL from API
    api_url = f"https://api.datamule.xyz/dataset/{urllib.parse.quote(dataset_name)}?api_key={api_key}"

    req = urllib.request.Request(
        api_url,
        headers={
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        }
    )
    
    try:
        with urllib.request.urlopen(req) as response:
            data = json.loads(response.read().decode())
    except urllib.error.HTTPError as e:
        error_body = e.read().decode()
        raise Exception(f"API request failed: {error_body}")
    
    if not data.get('success'):
        raise Exception(f"API error: {data.get('error', 'Unknown error')}")
    
    billing = data.get('metadata', {}).get('billing', {})
    cost = billing.get('total_charge', 0)
    remaining_balance = billing.get('remaining_balance')

    download_url = data['data']['download_url']
    size_gb = data['data']['size_gb']
    
    # Extract filename from URL if not provided
    if filename is None:
        parsed = urllib.parse.urlparse(download_url)
        query_params = urllib.parse.parse_qs(parsed.query)
        path = query_params.get('path', [''])[0]
        filename = urllib.parse.unquote(path.split('/')[-1])
        if not filename:
            filename = f"{dataset}.download"
    
    print(f"Downloading {dataset} ({size_gb:.2f} GB)...")
    
    download_req = urllib.request.Request(
        download_url,
        headers={
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        }
    )
    
    try:
        with urllib.request.urlopen(download_req) as response:
            total_size = int(response.headers.get('Content-Length', 0))
            
            with open(filename, 'wb') as f, tqdm(
                total=total_size,
                unit='B',
                unit_scale=True,
                desc=filename
            ) as pbar:
                while True:
                    chunk = response.read(8192)
                    if not chunk:
                        break
                    f.write(chunk)
                    pbar.update(len(chunk))
    except urllib.error.HTTPError as e:
        error_body = e.read().decode()
        raise Exception(f"Download failed: {error_body}")
    
    print(f"Downloaded to {filename}")
    if cost is not None:
        print(f"- Cost: ${cost:.4f} | Remaining balance: ${remaining_balance:.2f}")


def format_accession(accession, format):
    if format == 'int':
        accession = int(str(accession).replace('-',''))
    elif format == 'dash':
        accession = str(int(str(accession).replace('-',''))).zfill(18)
        accession = f"{accession[:10]}-{accession[10:12]}-{accession[12:]}"
    elif format == 'no-dash':
        accession = str(int(str(accession).replace('-',''))).zfill(18)
    else:
        raise ValueError("unrecognized format")
    return accession

def detect_accession_type(accession):
    accession = str(accession)
    if '-' in accession:
        return 'dash'
    elif len(accession) == 18:
        return 'no-dash'
    else:
        return 'int'


from setuptools import setup

setup(
    name="datamule_hub",
    author="John Friedman",
    version="0.0.1",
    description="Access Datamule cloud",
    url="https://github.com/john-friedman/datamule-hub",
    install_requires=[
        "tqdm",
        "aiohttp",
        "aioboto3",
        "gcloud-aio-storage",
    ],
)


import asyncio
import aiohttp
import aioboto3
import ssl
import json
from urllib.parse import urlparse
from tqdm import tqdm
from datetime import datetime, timezone

from ..utils import _generate_dates, _get_urls

async def _transfer_file(session, s3_client, semaphore, url, bucket, prefix=None, retry_errors=3):
    async with semaphore:
        key = urlparse(url).path.split('/')[-1]
        if prefix:
            key = f"{prefix.rstrip('/')}/{key}"
        last_error = None
        for attempt in range(retry_errors + 1):
            try:
                async with session.get(url) as response:
                    if response.status != 200:
                        raise aiohttp.ClientResponseError(response.request_info, response.history, status=response.status)
                    content = await response.read()
                    await s3_client.put_object(
                        Bucket=bucket, Key=key, Body=content,
                        ContentType=response.headers.get('Content-Type', 'application/octet-stream'),
                        Metadata={'source-url': url, 'transfer-date': datetime.now(timezone.utc).isoformat()}
                    )
                    return {'success': True, 'url': url, 'size_bytes': len(content)}
            except Exception as e:
                if attempt < retry_errors:
                    await asyncio.sleep(2 ** attempt)
                last_error = e
        return {'success': False, 'url': url, 'error': str(last_error)}


async def _transfer_urls(urls, s3_credentials, max_workers, retry_errors, prefix=None):
    connector = aiohttp.TCPConnector(limit=max_workers, ssl=ssl.create_default_context(), ttl_dns_cache=300)
    async with aiohttp.ClientSession(connector=connector, timeout=aiohttp.ClientTimeout(total=600)) as session:
        async with aioboto3.Session().client(
            's3',
            aws_access_key_id=s3_credentials['aws_access_key_id'],
            aws_secret_access_key=s3_credentials['aws_secret_access_key'],
            region_name=s3_credentials['region_name']
        ) as s3_client:
            semaphore = asyncio.Semaphore(max_workers)
            tasks = [_transfer_file(session, s3_client, semaphore, url, s3_credentials['bucket_name'], prefix, retry_errors) for url in urls]

            failed, total_bytes = [], 0
            with tqdm(total=len(urls), desc="Transferring files", unit="file") as pbar:
                for coro in asyncio.as_completed(tasks):
                    result = await coro
                    if result['success']:
                        total_bytes += result['size_bytes']
                    else:
                        failed.append(result)
                    pbar.set_postfix({'Total': f"{total_bytes / (1024**3):.2f} GB"})
                    pbar.update(1)

            return failed


def bucket_transfer(datamule_bucket, s3_credentials, max_workers=4, errors_json_filename='errors_bucket_transfer.json',
                retry_errors=3, force_daily=True, cik=None, submission_type=None, filing_date=None, accession_number=None, prefix=None):

    if datamule_bucket not in ['filings_sgml_r2', 'sec_filings_sgml_r2']:
        raise ValueError('Datamule S3 bucket not found.')

    if accession_number is not None and any(p is not None for p in [cik, submission_type, filing_date]):
        raise ValueError('If accession_number is provided, cik, submission_type, and filing_date must be None.')

    dates = _generate_dates(filing_date) if force_daily and filing_date else [filing_date]

    for date in dates:
        if len(dates) > 1:
            print(f"Transferring {date}")
        urls = _get_urls(submission_type=submission_type, cik=cik, filing_date=date, accession_number=accession_number)
        failed = asyncio.run(_transfer_urls(urls, s3_credentials, max_workers, retry_errors, prefix=prefix))

        if failed and errors_json_filename:
            with open(errors_json_filename, 'w') as f:
                json.dump(failed, f, indent=2)
            print(f"Saved {len(failed)} errors to {errors_json_filename}")

        print(f"Transfer complete: {len(urls) - len(failed)}/{len(urls)} files successful")